/*********************************************************************
 * webcam_server                                                     *
 *                                                                   *
 * (c) 2002 Donn Morrison donn@donn.dyndns.org                       *
 *                                                                   *
 * code used from Gerd Knorr's xawtv (libng)                         *
 * - and -                                                           *
 * Cory Lueninghoener's gqcam                                        *
 *                                                                   *
 *    waits for connections from a viewer and sends                  *
 *    jpeg encoded captures as a live video feed                     *
 *                                                                   *
 *********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <semaphore.h>
#include <jpeglib.h>
#include <jerror.h>
#include <semaphore.h>
#include <pthread.h>
#include <string.h>
#include <linux/videodev.h>

#include "webcam_server.h"
#include "filters.h"
#include "jdatabuf.h"

/*-----------------------------------------------------------
  image processing procedures
  
  raw_to_jpeg: encodes raw image to jpeg
  add_text: adds a caption to the image
  swap_rgb24: RGB->RBG
  fliph: flips the image horizontally
  flipv: flips vertically
  -----------------------------------------------------------*/
int raw_to_jpeg(char *raw_image, char **jpeg_data, int w, int h, int quality)
{
	struct jpeg_compress_struct cinfo;
	struct jpeg_error_mgr jerr;
	int i, size;
	unsigned char *line;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_compress(&cinfo);
	jpeg_buff_dest(&cinfo);
	cinfo.image_width = w;
	cinfo.image_height = h;
	cinfo.input_components = 3;
	cinfo.in_color_space = JCS_RGB;
	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE);
	jpeg_start_compress(&cinfo, TRUE);
	for (i = 0, line = raw_image; i < h; i++, line += w*3)
		jpeg_write_scanlines(&cinfo, &line, 1);
	jpeg_finish_compress(&cinfo);
	
	size = get_jpeg_size(&cinfo);
	
	/* this memory must be freed elsewhere */
	(*jpeg_data) = (char *)malloc(size);
	if(!(*jpeg_data))
	{
		jpeg_destroy_compress(&cinfo);
		jpeg_buff_free(&cinfo);
		return -1;
	}
	memcpy(*jpeg_data, ((my_dest_ptr)cinfo.dest)->buffer, size);

	jpeg_destroy_compress(&cinfo);
	jpeg_buff_free(&cinfo);
	
	return size;
}

/* Gerd Knorr */

#define CHAR_HEIGHT  11
#define CHAR_WIDTH   6
#define CHAR_START   4
#include "font_6x11.h"

void add_text(struct caminfo *cam, char *image, int width, int height)
{
	time_t      t;
	struct tm  *tm;
	char        *line,*ptr,prepend[32];
	int         i,x,y,f,len,size,numlines;

	if(cam->o.caption && strlen(cam->o.caption) < 1)
		return;

	pthread_mutex_lock(&cam->lock_info);
	sprintf(prepend, ", [%d] %d viewer", cam->s.id_newest_img, cam->s.num_users);
	if(cam->s.num_users != 1)
	strcat(prepend, "s");
	pthread_mutex_unlock(&cam->lock_info);

	size = strlen(cam->o.caption) +
	       strlen(prepend) +
	       128 +			/* 128 for strftime and null */
	       height/CHAR_HEIGHT;	/* total lines space in image */

	line = (char *)malloc(size);
	if(!line)
		return;

	time(&t);
	tm = localtime(&t);
	len = strftime(line,size-1,cam->o.caption,tm) + strlen(prepend);
	strcat(line,prepend);

	numlines = strlen(line) / (width/CHAR_WIDTH) + (strlen(line)%(width/CHAR_WIDTH) ? 1 : 0);

	for (y = 0; y < CHAR_HEIGHT; y++) {
		numlines = strlen(line) / (width/CHAR_WIDTH) + (strlen(line)%(width/CHAR_WIDTH) ? 1 : 0);
		ptr = image + 3 * width * (height-CHAR_HEIGHT-2 - ((numlines-1) * CHAR_HEIGHT) +y) + 12;
		for (x = 0; x < len; x++) {
			if(x % ((width/CHAR_WIDTH)-1) == 0 && x > 0)
			{
				numlines --;
				ptr = image + 3 * width * (height-CHAR_HEIGHT-2 - ((numlines-1) * CHAR_HEIGHT) +y) + 12;
			}
			f = fontdata[line[x] * CHAR_HEIGHT + y];
			for (i = CHAR_WIDTH-1; i >= 0; i--) {
				if (f & (CHAR_START << i)) {
					ptr[0] = cam->o.text_fg;
					ptr[1] = cam->o.text_fg;
					ptr[2] = cam->o.text_fg;
				}
				ptr += 3;
			}
		}
	}
	free(line);
}

void swap_rgb24(char *mem, int n)
{
        char  c;
        char *p = mem;
        int   i = n;

        while (--i) {
                c = p[0]; p[0] = p[2]; p[2] = c;
                p += 3;
        }
}

void adjust_gamma(char *mem, int n, int corr)
{
	int i=n;
	char *p = mem;
	while(--i)
	{
		if(corr > 0)
		{
			if(p[0] + corr <= 255)
				p[0] += corr;
			else
				p[0] = 255;
		}
		else
		{
			if(p[0] + corr >= 0)
				p[0] += corr;
			else
				p[0] = 0;
		}
		p++;
	}
}

unsigned char *
rotate_image(unsigned char * in, int *wp, int *hp, int rot,
	     int top, int left, int bottom, int right)
{
    static unsigned char * rotimg = NULL;

    int i, j;

    int w = *wp;
    int ow = (right-left);
    int oh = (bottom-top);

    if (rotimg == NULL && (rotimg = (char*)malloc(ow*oh*3)) == NULL ) {
	fprintf(stderr, "out of memory\n");
	exit(1);
    }
    switch ( rot ) {
    default:
    case 0:
	for (j = 0; j < oh; j++) {
	    int ir = (j+top)*w+left;
	    int or = j*ow;
	    for (i = 0; i < ow; i++) {
		rotimg[3*(or + i)]   = in[3*(ir+i)];
		rotimg[3*(or + i)+1] = in[3*(ir+i)+1];
		rotimg[3*(or + i)+2] = in[3*(ir+i)+2];
	    }
	}
	*wp = ow;
	*hp = oh;
	break;
    case 1:
	for (i = 0; i < ow; i++) {
	    int rr = (ow-1-i)*oh;
	    int ic = i+left;
	    for (j = 0; j < oh; j++) {
		rotimg[3*(rr+j)]   = in[3*((j+top)*w+ic)];
		rotimg[3*(rr+j)+1] = in[3*((j+top)*w+ic)+1];
		rotimg[3*(rr+j)+2] = in[3*((j+top)*w+ic)+2];
	    }
	}
	*wp = oh;
	*hp = ow;
	break;
    case 2:
	for (j = 0; j < oh; j++) {
	    int ir = (j+top)*w;
	    for (i = 0; i < ow; i++) {
		rotimg[3*((oh-1-j)*ow + (ow-1-i))] = in[3*(ir+i+left)];
		rotimg[3*((oh-1-j)*ow + (ow-1-i))+1] = in[3*(ir+i+left)+1];
		rotimg[3*((oh-1-j)*ow + (ow-1-i))+2] = in[3*(ir+i+left)+2];
	    }
	}
	*wp = ow;
	*hp = oh;
	break;
    case 3:
	for (i = 0; i < ow; i++) {
	    int rr = i*oh;
	    int ic = i+left;
	    rr += oh-1;
	    for (j = 0; j < oh; j++) {
		rotimg[3*(rr-j)]   = in[3*((j+top)*w+ic)];
		rotimg[3*(rr-j)+1] = in[3*((j+top)*w+ic)+1];
		rotimg[3*(rr-j)+2] = in[3*((j+top)*w+ic)+2];
	    }
	}
	*wp = oh;
	*hp = ow;
	break;
    }
    memcpy(in, rotimg, ow*oh*3);
    return rotimg;
}

unsigned int
compare_images(unsigned char *last, unsigned char *current,
	       int width, int height)
{
    unsigned char *p1 = last;
    unsigned char *p2 = current;
    int avg, diff, max, i = width*height*3;

    for (max = 0, avg = 0; --i; p1++,p2++) {
	diff = (*p1 < *p2) ? (*p2 - *p1) : (*p1 - *p2);
	avg += diff;
	if (diff > max)
	    max = diff;
    }
    avg = avg / width / height;
    fprintf(stderr,"compare: max=%d,avg=%d\n",max,avg);
    /* return avg */
    return max;
}

void fliph(char *image, int width, int height)
{
	int i,j;
	char *dest = NULL;
	char *srcline, *destline;

	dest = (char *)malloc(width*height*3);
	if(!dest)
		return;

	for(i=0;i<height;i++)
	{
		srcline = image + i*3*width;
		destline = dest + i*3*width + 3*(width-1);
		for(j=0;j<width;j++)
		{
			memcpy(destline, srcline, 3);
			srcline += 3;
			destline -= 3;
		}
	}

	memcpy(image, dest, width*height*3);
	free(dest);
}

void flipv(char *image, int width, int height)
{
	int i;
	char *dest = NULL;
	char *srcline, *destline;
	
	dest = (char *)malloc(width*height*3);
	if(!dest)
		return;
	
	for(i=0;i<height;i++)
	{
		srcline = image + i*width*3;
		destline = dest + (height-i)*width*3 - width*3;
	//	printf("%d - %d = %d\n", height*width*3, (height-i)*width*3 - width*3, height*width*3 - ((height-i)*width*3 - width*3));
		memcpy(destline, srcline, width*3);
	}
	
	memcpy(image, dest, width*height*3);
	free(dest);
}
